﻿/****************************************************************************

 ■HiResAmp 全体制御(ControlApp)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

using System;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Configuration;

namespace ControlApp
{
    public partial class MainForm : Form
    {
        private readonly string[] StateNames = {
            "",
            "開始中",
            "▶再生中",
            "切り替え中",
            "停止中",
        };

        private string m_ScriptFile;
        private Player m_Player;
        private Point m_LastPos;
        private int m_LastKeyHook;
        private Color m_SaveColor;
        private string m_TitleBase;
        private bool m_Modified;

        //---------------------------------------------------------------

        /// <summary>
        /// ●コンストラクタ
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            try
            {
                string path = System.AppDomain.CurrentDomain.BaseDirectory;

                if (!File.Exists(Path.Combine(path, "ControlApp.exe.config")))
                    throw new Exception("ControlApp.exe.config ファイルがありません");

                if (!File.Exists(Path.Combine(path, "Settings.txt")))
                    throw new Exception("Settings.txt ファイルがありません");

                m_ScriptFile = Path.Combine(path, "Settings.txt");

                m_Player = new Player(OnStateChange, OnPlayerInstruction, OnPlayerException);

                m_LastPos = Point.Empty;
                m_Modified = false;
                m_LastKeyHook = 0;
            }
            catch (Exception ex)
            {
                WindowState = FormWindowState.Normal;
                ShowException(ex);
                Close();
            }
        }

        /// <summary>
        /// ●ロード
        /// </summary>
        private void MainForm_Load(object sender, EventArgs e)
        {
            m_TopMost.Checked = TopMost;

            // 画面の解像度を表示
            WinApi.RECT rect;
            WinApi.GetWindowRect(WinApi.GetDesktopWindow(), out rect);
            m_TitleBase = string.Format("{0}　{1}x{2}", Text, rect.right, rect.bottom);
            Text = m_TitleBase;

            // テキストボックスのタブサイズを設定
            int[] arg = new int[] { 4 * 4 };
            IntPtr ptr = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(int)) * arg.Length);
            Marshal.Copy(arg, 0, ptr, arg.Length);
            WinApi.SendMessage(m_ScriptEdit.Handle, 203, (UIntPtr)1, (UIntPtr)ptr.ToInt64());
            Marshal.FreeCoTaskMem(ptr);

            m_SerialPort.NewLine = "\r\n";
            m_PortName.Text = ConfigurationManager.AppSettings["PortName"];

            try
            {
                // ファイルをロード
                using (StreamReader sr = new StreamReader(m_ScriptFile))
                {
                    m_ScriptEdit.Text = sr.ReadToEnd();
                    m_ScriptEdit.SelectionStart = 0;
                }

                m_Save.Enabled = false;
                m_Modified = false;

                // 初期化
                Initialize();

                // 起動時の自動開始
                if (m_Player.IsEnable)
                {
                    string[] args = Environment.GetCommandLineArgs();
                    if (args.Length < 2 || args[1] != "/D")
                        m_Player.Start();
                }
            }
            catch (Exception ex)
            {
                m_Player.Stop();

                if (m_SerialPort.IsOpen)
                {
                    try { m_SerialPort.WriteLine("ERR"); }
                    catch { }
                }

                WindowState = FormWindowState.Normal;
                ShowException(ex);
            }
        }

        /// <summary>
        /// ●初期化
        /// </summary>
        private void Initialize()
        {
            m_Log.Items.Clear();
            m_Err.Text = "";

            if (m_SerialPort.IsOpen)
                m_SerialPort.Close();

            m_SerialPort.PortName = m_PortName.Text;
            m_SerialPort.Open();

            try
            {
                m_Player.Load(m_ScriptEdit.Text);
            }
            catch(Exception ex)
            {
                OnPlayerException(ex);
                return;
            }

            m_SerialPort.WriteLine("STARTED");
        }

        /// <summary>
        /// ●設定保存
        /// </summary>
        private void m_Save_Click(object sender, EventArgs e)
        {
            // 保存
            try
            {
                using (StreamWriter sw = File.CreateText(m_ScriptFile))
                {
                    sw.Write(m_ScriptEdit.Text);
                }
                m_Save.Enabled = false;
                m_Modified = false;
            }
            catch (Exception ex)
            {
                ShowException(ex);
                return;
            }
        }

        /// <summary>
        /// ●説明
        /// </summary>
        private void m_Help_Click(object sender, EventArgs e)
        {
            HelpDlg dlg = new HelpDlg();
            dlg.TopMost = this.TopMost;
            dlg.ShowDialog();
        }

        /// <summary>
        /// ●シリアル受信
        /// </summary>
        private void m_SerialPort_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            try {
                OnReceive(m_SerialPort.ReadLine());
            }
            catch { }
        }

        /// <summary>
        /// ●シリアル受信
        /// </summary>
        private void OnReceive(string text)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new Action<string>(OnReceive), new object[] { text });
                return;
            }
            try
            {
                switch (text)
                {
                case "PREV1":
                    m_SerialPort.WriteLine("OK");
                    m_Player.PrevOrder();
                    break;

                case "PREV2":
                    m_SerialPort.WriteLine("OK");
                    m_Player.PrevGenre();
                    break;

                case "NEXT1":
                    m_SerialPort.WriteLine("OK");
                    m_Player.NextOrder();
                    break;

                case "NEXT2":
                    m_SerialPort.WriteLine("OK");
                    m_Player.NextGenre();
                    break;

                case "SHUTDOWN":
                    m_SerialPort.WriteLine("OK");
                    m_Player.Stop();

                    Thread.Sleep(100);
                    m_SerialPort.Close();

                    // シャットダウン
                    WinApi.AdjustToken();
                    WinApi.ExitWindowsEx(WinApi.ExitWindows.EWX_POWEROFF | WinApi.ExitWindows.EWX_FORCE, 0);

                    Close();
                    break;
                }
            }
            catch (Exception ex)
            {
                m_Player.Stop();

                try { m_SerialPort.WriteLine("ERR"); }
                catch { }

                ShowException(ex);
            }
        }

        /// <summary>
        /// ●プレイヤー状態変化
        /// </summary>
        private void OnStateChange(Player.ProcState state)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new Action<Player.ProcState>(OnStateChange), state);
                return;
            }

            if (state != Player.ProcState.Stopped)
                Text = string.Format("{0}　【{1}】", m_TitleBase, StateNames[(int)state]);
            else
                Text = m_TitleBase;

            m_PrevGenre.Enabled = (state == Player.ProcState.Playing);
            m_NextGenre.Enabled = m_PrevGenre.Enabled;
            m_PrevOrder.Enabled = m_PrevGenre.Enabled;
            m_NextOrder.Enabled = m_PrevGenre.Enabled;
        }

        /// <summary>
        /// ●プレイヤー実行情報
        /// </summary>
        private void OnPlayerInstruction(int line, string info)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new Action<int, string>(OnPlayerInstruction), line, info);
                return;
            }

            bool scroll = false;
            if (line > 0)
            {
                // 最後までスクロールしていたら追加後に自動スクロールする
                if (m_Log.Items.Count > 0)
                {
                    int hight = m_Log.Items[0].Bounds.Height;
                    int num = (m_Log.ClientRectangle.Height + hight - 1) / hight;

                    if ((m_Log.TopItem.Index + num) >= m_Log.Items.Count)
                        scroll = true;
                }

                // 500行を越えた分は古い方から削除
                if (m_Log.Items.Count >= 500)
                    m_Log.Items.RemoveAt(0);
            }
            if (line >= 0)
            {
                m_Log.Items.Add(line.ToString()).SubItems.Add(info);

                if (scroll)
                    m_Log.EnsureVisible(m_Log.Items.Count - 1);

                m_Log.Update();
            }
            else
            {
                m_Err.Text = info;
                m_Err.Update();
            }
        }

        /// <summary>
        /// ●プレイヤー例外
        /// </summary>
        private void OnPlayerException(Exception ex)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new Action<Exception>(OnPlayerException), ex);
                return;
            }

            try { m_SerialPort.WriteLine("ERR"); }
            catch { }

            if (ex is PlayerException)
                m_Err.Text = ex.Message;
            else
                m_Err.Text = ex.Message + "\r\n" + ex.StackTrace;
        }

        /// <summary>
        /// ●クリック記述
        /// </summary>
        private void m_Click_Click(object sender, EventArgs e)
        {
            if (!m_Click.Checked)
            {
                m_Player.SetForegroundWindow();

                MouseHook.StartHook(OnMouseHook);
                KeyHook.StartHook(OnKeyHook);
                m_Click.Checked = true;
                m_Pixel.Checked = false;
                m_Ifgo.Checked = false;
                m_SaveColor = BackColor;
                BackColor = Color.MediumAquamarine;
            }
        }

        /// <summary>
        /// ●色待ち記述
        /// </summary>
        private void m_Pixel_Click(object sender, EventArgs e)
        {
            if (!m_Pixel.Checked)
            {
                m_Player.SetForegroundWindow();

                MouseHook.StartHook(OnMouseHook);
                KeyHook.StartHook(OnKeyHook);
                m_Click.Checked = false;
                m_Pixel.Checked = true;
                m_Ifgo.Checked = false;
                m_SaveColor = BackColor;
                BackColor = Color.MediumAquamarine;
            }
        }

        /// <summary>
        /// ●条件G記述
        /// </summary>
        private void m_Ifgo_Click(object sender, EventArgs e)
        {
            if (!m_Ifgo.Checked)
            {
                m_Player.SetForegroundWindow();

                MouseHook.StartHook(OnMouseHook);
                KeyHook.StartHook(OnKeyHook);
                m_Click.Checked = false;
                m_Pixel.Checked = false;
                m_Ifgo.Checked = true;
                m_SaveColor = BackColor;
                BackColor = Color.MediumAquamarine;
            }
        }

        /// <summary>
        /// ●開始/停止
        /// </summary>
        private void m_Stop_Click(object sender, EventArgs e)
        {
            if (m_Player.State == Player.ProcState.Stopped)
            {
                try
                {
                    // 初期化
                    Initialize();

                    // 開始
                    if (m_Player.IsEnable)
                        m_Player.Start();
                }
                catch (Exception ex)
                {
                    m_Player.Stop();

                    if (m_SerialPort.IsOpen)
                    {
                        try { m_SerialPort.WriteLine("ERR"); }
                        catch { }
                    }

                    ShowException(ex);
                }
            }
            else
            {
                m_Player.Stop();
            }
        }

        /// <summary>
        /// ●マウスフックハンドラ
        /// </summary>
        private bool OnMouseHook(UIntPtr wParam, Point pt)
        {
            try
            {
                if ((uint)wParam == MouseHook.WM_MOUSEMOVE)
                {
                    if (pt != m_LastPos)
                    {
                        m_LastPos = pt;

                        // カーソル周辺画像取得
                        Color center = new Color();
                        using (Bitmap src = new Bitmap(11, 11))
                        {
                            using (Graphics g = Graphics.FromImage(src))
                            {
                                Point org = pt;
                                org.Offset(-5, -5);
                                g.CopyFromScreen(org, Point.Empty, src.Size);
                            }

                            // カーソル周辺拡大画像表示
                            Bitmap bmp = new Bitmap(110, 110);

                            using (Graphics g = Graphics.FromImage(bmp))
                            {
                                for (int x = 0; x < 11; x++)
                                {
                                    for (int y = 0; y < 11; y++)
                                    {
                                        Color col = src.GetPixel(x, y);

                                        if (x == 5 && y == 5)
                                            center = col;

                                        SolidBrush brs = new SolidBrush(col);
                                        Rectangle rect = new Rectangle(x * 10, y * 10, 10, 10);
                                        g.FillRectangle(brs, rect);
                                    }
                                }
                                g.DrawRectangle(new Pen(Color.Red), new Rectangle(50, 50, 9, 9));

                                m_Zoom.Image = bmp;
                                m_Zoom.Update();
                            }
                        }

                        // 座標表示
                        pt = m_Player.ToWindow(pt);
                        m_Position.Text = string.Format("X:{0} Y:{1}", pt.X, pt.Y);
                        m_Color.Text = string.Format("#{0:X2}{1:X2}{2:X2}", center.R, center.G, center.B);
                        m_Position.Update();
                        m_Color.Update();
                    }

                    return false;
                }
                else if ((UInt32)wParam == MouseHook.WM_LBUTTONDOWN)
                {
                    GetDescription();
                    StopMouseSetting();

                    return true;
                }
                else if ((UInt32)wParam == MouseHook.WM_RBUTTONDOWN)
                {
                    StopMouseSetting();

                    return true;
                }
            }
            catch (Exception ex)
            {
                StopMouseSetting();
                ShowException(ex);
            }

            return false;
        }

        /// <summary>
        /// ●キーフックハンドラ
        /// </summary>
        private bool OnKeyHook(UIntPtr wParam, Keys key)
        {
            if (KeyHook.WM_KEYDOWN == (UInt32)wParam)
            {
                if (key == Keys.Escape)
                {
                    StopMouseSetting();
                    return true;
                }
                else if (key == Keys.Left || key == Keys.Up || key == Keys.Right || key == Keys.Down)
                {
                    if ((Environment.TickCount - m_LastKeyHook) >= 10)
                    {
                        Point pt = Cursor.Position;
                        switch (key)
                        {
                        case Keys.Left: --pt.X; break;
                        case Keys.Right: ++pt.X; break;
                        case Keys.Up: --pt.Y; break;
                        case Keys.Down: ++pt.Y; break;
                        }
                        Cursor.Position = pt;
                        Thread.Sleep(100); // マウスホバーによる色の変化を待つ

                        OnMouseHook((UIntPtr)MouseHook.WM_MOUSEMOVE, Cursor.Position);
                        m_LastKeyHook = Environment.TickCount;
                    }

                    return true;
                }
            }

            return false;
        }

        /// <summary>
        /// ●記述を取得する
        /// </summary>
        private void GetDescription()
        {
            if (m_Click.Checked)
            {
                // クリック記述
                Point pt = m_Player.ToWindow(Cursor.Position);
                Clipboard.SetText(string.Format("Click({0},{1})", pt.X, pt.Y));
            }
            else
            {
                using (Bitmap bmp = new Bitmap(1, 1))
                {
                    using (Graphics g = Graphics.FromImage(bmp))
                    {
                        Point pt = Cursor.Position;
                        g.CopyFromScreen(pt, Point.Empty, bmp.Size);
                        Color col = bmp.GetPixel(0, 0);

                        pt = m_Player.ToWindow(pt);

                        if (m_Pixel.Checked)
                        {
                            // 色待ち記述
                            Clipboard.SetText(
                                string.Format("WaitPixel({0},{1},#{2:X2}{3:X2}{4:X2},10000)",
                                    pt.X, pt.Y, col.R, col.G, col.B));
                        }
                        else
                        {
                            // 条件G記述
                            Clipboard.SetText(
                                string.Format("IfGoto({0},{1},#{2:X2}{3:X2}{4:X2},Label)",
                                    pt.X, pt.Y, col.R, col.G, col.B));
                        }
                    }
                }
            }
        }

        /// <summary>
        /// ●前のジャンル
        /// </summary>
        private void m_PrevGenre_Click(object sender, EventArgs e)
        {
            m_Player.PrevGenre();
        }

        /// <summary>
        /// ●次のジャンル
        /// </summary>
        private void m_NextGenre_Click(object sender, EventArgs e)
        {
            m_Player.NextGenre();
        }

        /// <summary>
        /// ●前のプレイリスト
        /// </summary>
        private void m_PrevOrder_Click(object sender, EventArgs e)
        {
            m_Player.PrevOrder();
        }

        /// <summary>
        /// ●次のプレイリスト
        /// </summary>
        private void m_NextOrder_Click(object sender, EventArgs e)
        {
            m_Player.NextOrder();
        }

        /// <summary>
        /// ●TopMost
        /// </summary>
        private void m_TopMost_CheckedChanged(object sender, EventArgs e)
        {
            TopMost = m_TopMost.Checked;
        }

        /// <summary>
        /// ●カーソル移動
        /// </summary>
        private void m_MoveCursor_CheckedChanged(object sender, EventArgs e)
        {
            m_Player.CursorMove = m_CursorMove.Checked;
        }

        /// <summary>
        /// ●実行リストクリック
        /// </summary>
        private void m_Log_Click(object sender, EventArgs e)
        {
            if (m_Log.SelectedItems.Count > 0)
            {
                int line = int.Parse(m_Log.SelectedItems[0].Text);
                if (line == 0)
                {
                    m_ScriptEdit.Select(0, 0);
                    m_ScriptEdit.ScrollToCaret();
                    return;
                }

                string text = m_ScriptEdit.Text;
                if (text != "")
                {
                    int cur = 1, start = 0, end = 0;
                    for (int i = 0; i < text.Length; i++)
                    {
                        if (text[i] == '\n')
                        {
                            end = i - 2;

                            if (cur == line)
                            {
                                m_ScriptEdit.Select(start, end - start + 1);
                                m_ScriptEdit.ScrollToCaret();
                                break;
                            }

                            if ((i + 1) >= text.Length)
                            {
                                m_ScriptEdit.Select(text.Length, text.Length);
                                m_ScriptEdit.ScrollToCaret();
                                break;
                            }

                            cur++;
                            start = i + 1;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// ●スクリプト編集
        /// </summary>
        private void m_ScriptEdit_TextChanged(object sender, EventArgs e)
        {
            m_Save.Enabled = true;
            m_Modified = true;
        }

        /// <summary>
        /// ●例外メッセージを表示する
        /// </summary>
        private void ShowException(Exception ex)
        {
            MessageBox.Show(this, ex.Message + "\r\n\r\n" + ex.StackTrace,
                this.Text, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        }

        /// <summary>
        /// ●フォーム最小化
        /// </summary>
        private void MainForm_SizeChanged(object sender, EventArgs e)
        {
            if (WindowState == FormWindowState.Minimized)
                StopMouseSetting();
        }

        /// <summary>
        /// ●フォームを閉じる
        /// </summary>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            Configuration config = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
            config.AppSettings.Settings["PortName"].Value = m_PortName.Text;
            config.Save();

            if (m_Modified)
            {
                if (MessageBox.Show(this, "スクリプトが編集されています。\n保存しますか？", "",
                    MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
                {
                    m_Save_Click(null, null);
                }
            }

            m_Player.Abort();
            m_SerialPort.Close();
            MouseHook.EndHook();
            KeyHook.EndHook();
        }

        /// <summary>
        /// ●マウス設定を停止する
        /// </summary>
        private void StopMouseSetting()
        {
            MouseHook.EndHook();
            KeyHook.EndHook();

            BackColor = m_SaveColor;

            m_Click.Checked = false;
            m_Pixel.Checked = false;
            m_Ifgo.Checked = false;
            m_Zoom.Image = null;
            m_Position.Text = string.Empty;
            m_Color.Text = string.Empty;
            m_LastPos = Point.Empty;
        }
    }
}
