﻿/****************************************************************************

 ■HiResAmp 全体制御(ControlApp)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

using System;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Configuration;
using System.Threading;

namespace ControlApp
{
    class Player
    {
        #region Public data

        public enum ProcState
        {
            Stopped,    // 停止
            Starting,   // 開始中
            Playing,    // 再生中
            Changing,   // 切り替え中
            Stopping,   // 停止中
        }

        public ProcState State { get; set; }
        public bool CursorMove { get; set; }
        public bool IsEnable { get; private set; }

        #endregion

        #region Private data

        private enum ChangeType
        {
            None,       // 標準(NextOrder)
            PrevOrder,  // 前のプレイリスト
            NextOrder,  // 次のプレイリスト
            PrevGenre,  // 前のジャンル
            NextGenre,  // 次のジャンル
        }

        private class PlayList
        {
            public int Order { get; set; }
            public int Genre { get; set; }
        }

        private Action<ProcState> m_OnStateChange;
        private Action<int, string> m_OnInstruction;
        private Action<Exception> m_OnException;

        private IntPtr m_AmazonMusic;
        private IntPtr m_OpeWnd;

        private List<PlayList> m_PlayList;

        private int m_CurOrder;
        private int m_CurGenre;

        private Thread m_Script;
        private bool m_Break;
        private ChangeType m_Change;

        private class ScrWork
        {
            public enum SystemLabel
            {
                None,
                Start,
                WaitEnd,
                Change,
                Stop,
                PlayList,
            }
            public enum LineType
            {
                Empty,
                SystemLabel,
                Label,
                Function,
                PlayList
            }
            public enum FuncType
            {
                WaitPixel,
                IfGoto,
                Click,
                Move,
                Wait,
                Goto,
                Let,
                Log,
                Exit
            }
            public enum TokenType
            {
                Num,    Var,    Sys,
                Add,    Sub,    Mul,
                Div,    Mod,    Lpr,
                Rpr,    End
            }

            public class LineData
            {
                //public SystemLabel Section;
                public int Index;
                public LineType Type;
                public FuncType Func;
                public string[] Args;
                public string Text;
            }
                
            public Regex RegSysLabel;
            public Regex RegLabel;
            public Regex RegFunc;
            public Regex RegLet;
            public Regex RegPlayList;
            public Regex RegColor;
            public Regex RegToken;

            public Dictionary<SystemLabel, int> EntryPoints;
            public Dictionary<string, int> LabelPoints;
            public Dictionary<string, int> Vars;
            public List<LineData> LineDatas;
            public SystemLabel Section;
            public int Instruction;
            public int Count;

            public Match Tokenizer;
            public TokenType Token;
            public int Value;

            public ScrWork()
            {
                RegSysLabel = new Regex(@"^::(\w+)$", RegexOptions.Compiled);
                RegLabel = new Regex(@"^:(\w+)$", RegexOptions.Compiled);
                RegFunc = new Regex(@"^(\w+)\((.*)\)$", RegexOptions.Compiled);
                RegLet = new Regex(@"^@(\w+)\s*=(.+)$", RegexOptions.Compiled);
                RegPlayList = new Regex(@"^(\d{2})-G(\d{1}):\s*(\d{1})", RegexOptions.Compiled);
                RegColor = new Regex(@"^([#!])([0-9a-fA-F]{6})$", RegexOptions.Compiled);
                RegToken = new Regex(@"(\d+)|(@\w+)|(\$\w+)|(\s+)|([\+\-\*\/\%\(\)])|($)", RegexOptions.Compiled);

                EntryPoints = new Dictionary<SystemLabel, int>();
                LabelPoints = new Dictionary<string, int>();
                Vars = new Dictionary<string, int>();
                LineDatas = new List<LineData>();
            }
        }

        private ScrWork m_ScrWork;

        #endregion

        //---------------------------------------------------------------

        /// <summary>
        /// ●コンストラクタ
        /// </summary>
        internal Player(Action<ProcState> sc, Action<int,string> ip, Action<Exception> ex)
        {
            State = ProcState.Stopped;

            m_OnStateChange = sc;
            m_OnInstruction = ip;
            m_OnException = ex;

            int.TryParse(ConfigurationManager.AppSettings["LastOrder"], out m_CurOrder);
            int.TryParse(ConfigurationManager.AppSettings["LastGenre"], out m_CurGenre);
        }

        /// <summary>
        /// ●スクリプトロード
        /// </summary>
        internal void Load(string text)
        {
            if (m_Break)
                return;

            Debug.Assert(State == ProcState.Stopped);

            IsEnable = false;
            m_PlayList = new List<PlayList>();
            m_ScrWork = new ScrWork();

            ScrWork.SystemLabel section = ScrWork.SystemLabel.None;

            // ラインループ
            string[] lines = text.Split(new string[] { "\r\n" }, StringSplitOptions.None);
            for (int i = 0; i < lines.Length; i++)
            {
                ScrWork.LineData data = new ScrWork.LineData();
                data.Index = i;

                string line = lines[i].Trim();
                data.Text = line;

                // コメント除去
                int comment = line.IndexOf("//");
                if (comment != -1)
                    line = line.Substring(0, comment).Trim();

                if (line == "")
                {
                    // 空白行
                    data.Type = ScrWork.LineType.Empty;
                    m_ScrWork.LineDatas.Add(data);
                    continue;
                }

                Match m = m_ScrWork.RegSysLabel.Match(line);
                if (m.Success)
                {
                    // システムラベル
                    ScrWork.SystemLabel label;
                    try {
                        label = (ScrWork.SystemLabel)Enum.Parse(typeof(ScrWork.SystemLabel), m.Groups[1].Value);
                    }
                    catch {
                        throw new PlayerException("スクリプト記述に誤りがあります：" + line);
                    }
                    if (section != label)
                    {
                        if (label != ScrWork.SystemLabel.PlayList)
                            m_ScrWork.EntryPoints[label] = i;

                        section = label;
                    }
                    data.Type = ScrWork.LineType.SystemLabel;
                    m_ScrWork.LineDatas.Add(data);
                    continue;
                }

                if (section != ScrWork.SystemLabel.PlayList)
                {
                    // スクリプト
                    m = m_ScrWork.RegLabel.Match(line);
                    if (m.Success)
                    {
                        // ラベル
                        m_ScrWork.LabelPoints[m.Groups[1].Value] = i;

                        data.Type = ScrWork.LineType.Label;
                        m_ScrWork.LineDatas.Add(data);
                        continue;
                    }

                    m = m_ScrWork.RegFunc.Match(line);
                    if (m.Success)
                    {
                        // 関数呼び出し
                        data.Type = ScrWork.LineType.Function;
                        try {
                            data.Func = (ScrWork.FuncType)Enum.Parse(typeof(ScrWork.FuncType), m.Groups[1].Value);
                        }
                        catch {
                            throw new PlayerException("スクリプト記述に誤りがあります：" + line);
                        }
                        data.Args = m.Groups[2].Value.Split(new char[] { ',' });
                        for (int j = 0; j < data.Args.Length; j++)
                            data.Args[j] = data.Args[j].Trim();

                        m_ScrWork.LineDatas.Add(data);
                        continue;
                    }

                    m = m_ScrWork.RegLet.Match(line);
                    if (m.Success)
                    {
                        // 代入
                        data.Type = ScrWork.LineType.Function;
                        data.Func = ScrWork.FuncType.Let;

                        data.Args = new string[] { m.Groups[1].Value.Trim(), m.Groups[2].Value.Trim() };

                        m_ScrWork.LineDatas.Add(data);
                        continue;
                    }

                    throw new PlayerException("スクリプト記述に誤りがあります：" + line);
                }
                else
                {
                    // レイリストと有効無効設定
                    m = m_ScrWork.RegPlayList.Match(line);
                    if (m.Success)
                    {
                        if (m.Groups[3].Value != "0")
                        {
                            PlayList pl = new PlayList();
                            pl.Order = int.Parse(m.Groups[1].Value);
                            pl.Genre = int.Parse(m.Groups[2].Value);
                            m_PlayList.Add(pl);
                        }

                        data.Type = ScrWork.LineType.PlayList;
                        m_ScrWork.LineDatas.Add(data);
                        continue;
                    }

                    throw new PlayerException("プレイリスト設定に誤りがあります：" + line);
                }
            }

            // チェック
            if (m_ScrWork.EntryPoints.Count != 4)
                throw new PlayerException("一部のシステムラベルが未定義です");
            if (m_PlayList.Count == 0)
                throw new PlayerException("有効なプレイリストがありません");

            IsEnable = true;
        }

        /// <summary>
        /// ●開始
        /// </summary>
        internal void Start()
        {
            if (IsEnable && !m_Break)
            {
                m_Script = new Thread(new ThreadStart(ScriptEntry));
                m_Script.Start();
            }
        }

        /// <summary>
        /// ●停止
        /// </summary>
        internal void Stop()
        {
            if (m_Script != null && !m_Break)
            {
                m_Break = true;
                while (m_Script.IsAlive)
                {
                    Thread.Sleep(1);
                    Application.DoEvents();
                }
                m_Script = null;
                m_Break = false;
            }
        }

        /// <summary>
        /// ●強制終了
        /// </summary>
        internal void Abort()
        {
            if (m_Script != null)
            {
                m_Script.Abort();
                m_Script = null;
            }
        }

        /// <summary>
        /// ◎前のプレイリスト
        /// </summary>
        internal void PrevOrder()
        {
            if (State == ProcState.Playing && m_Change == ChangeType.None)
                m_Change = ChangeType.PrevOrder;
        }

        /// <summary>
        /// ◎次のプレイリスト
        /// </summary>
        internal void NextOrder()
        {
            if (State == ProcState.Playing && m_Change == ChangeType.None)
                m_Change = ChangeType.NextOrder;
        }

        /// <summary>
        /// ◎前のジャンル
        /// </summary>
        internal void PrevGenre()
        {
            if (State == ProcState.Playing && m_Change == ChangeType.None)
                m_Change = ChangeType.PrevGenre;
        }

        /// <summary>
        /// ◎次のジャンル
        /// </summary>
        internal void NextGenre()
        {
            if (State == ProcState.Playing && m_Change == ChangeType.None)
                m_Change = ChangeType.NextGenre;
        }

        /// <summary>
        /// ●AmazonMusicを手前に表示する
        /// </summary>
        internal void SetForegroundWindow()
        {
            GetOpeWnd();
            WinApi.SetForegroundWindow(m_AmazonMusic);
        }

        /// <summary>
        /// ●スクリーン座標を操作ウィンドウ座標へ変換する
        /// </summary>
        internal Point ToWindow(Point pt)
        {
            WinApi.POINT temp = pt;
            WinApi.MapWindowPoints(WinApi.GetDesktopWindow(), GetOpeWnd(), ref temp, 1);
            return temp;
        }

        /// <summary>
        /// ●操作ウィンドウ座標をスクリーン座標へ変換する
        /// </summary>
        internal Point ToScreen(Point pt)
        {
            WinApi.POINT temp = pt;
            WinApi.MapWindowPoints(GetOpeWnd(), WinApi.GetDesktopWindow(), ref temp, 1);
            return temp;
        }

        //---------------------------------------------------------------

        /// <summary>
        /// ●スレッドエントリ
        /// </summary>
        private void ScriptEntry()
        {
            try
            {
                ScriptProc();
                SetState(ProcState.Stopped);
            }
            catch (Exception ex)
            {
                SetState(ProcState.Stopped);
                m_OnException(ex);
            }
        }

        /// <summary>
        /// ●スクリプトスレッド
        /// </summary>
        private void ScriptProc()
        {
            SetState(ProcState.Starting);
            m_OnInstruction(0, "Looking for Amazon Music ...");

            // AmazonMusicを探す
            for (int i = 0; i < 600; i++) // タイムアウト1分
            {
                if (FindAmazonMusic())
                {
                    WinApi.SetForegroundWindow(m_AmazonMusic);
                    break;
                }
                if (!m_Break)
                {
                    Thread.Sleep(100);
                    continue;
                }
                return;
            }

            if (m_AmazonMusic == IntPtr.Zero)
                throw new PlayerException("Amazon Music が起動していません。");
            if (m_OpeWnd == IntPtr.Zero)
                throw new PlayerException("Amazon Music 操作ウィンドウが見つかりません。");

            // 開始スクリプト実行
            ScriptExecute(ScrWork.SystemLabel.Start);

            while (!m_Break)
            {
                m_Change = ChangeType.None;
                SetState(ProcState.Playing);

                // プレイリスト終了待ちスクリプト実行
                ScriptExecute(ScrWork.SystemLabel.WaitEnd);
                if (m_Break)
                    break;

                SetState(ProcState.Changing);

                switch (m_Change)
                {
                case ChangeType.PrevOrder:  // 現在のジャンルの前のプレイリスト
                    SetupNextOrder(ChangeType.PrevOrder);
                    break;
                case ChangeType.NextOrder:  // 現在のジャンルの次のプレイリスト
                    SetupNextOrder(ChangeType.NextOrder);
                    break;
                case ChangeType.PrevGenre:  // 前のジャンルの最初のプレイリスト
                    SetupNextOrder(ChangeType.PrevGenre);
                    break;
                case ChangeType.NextGenre:  // 次のジャンルの最初のプレイリスト
                    SetupNextOrder(ChangeType.NextGenre);
                    break;
                default: // プレイリスト通常終了→次のプレイリストへ
                    SetupNextOrder(ChangeType.None);
                    break;
                }

                // プレイリスト変更スクリプト実行
                ScriptExecute(ScrWork.SystemLabel.Change);
            }

            SetState(ProcState.Stopping);

            // 停止スクリプト実行
            ScriptExecute(ScrWork.SystemLabel.Stop);
        }

        /// <summary>
        /// ●スクリプト実行処理
        /// </summary>
        private void ScriptExecute(ScrWork.SystemLabel label)
        {
            m_ScrWork.Section = label;
            m_ScrWork.Instruction = m_ScrWork.EntryPoints[label];
            m_ScrWork.Count = 0;

            bool isbreak = false;
            bool complete = false;
            int lastinst = 0;
            try
            {
                for (; !isbreak && !complete; isbreak = IsBreak())
                {
                    ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

                    if (data.Type != ScrWork.LineType.Empty)
                    {
                        if (m_ScrWork.Count == 0 || data.Type != ScrWork.LineType.SystemLabel)
                            m_OnInstruction(m_ScrWork.Instruction + 1, data.Text);
                    }

                    switch (data.Type)
                    {
                    case ScrWork.LineType.Function:
                        lastinst = m_ScrWork.Instruction;
                        switch (data.Func)
                        {
                        case ScrWork.FuncType.WaitPixel:
                            FuncWaitPixel();
                            break;
                        case ScrWork.FuncType.IfGoto:
                            m_ScrWork.Instruction = FuncIfGoto(m_ScrWork.Instruction);
                            break;
                        case ScrWork.FuncType.Click:
                            FuncClick();
                            break;
                        case ScrWork.FuncType.Move:
                            FuncMove();
                            break;
                        case ScrWork.FuncType.Wait:
                            FuncWait();
                            break;
                        case ScrWork.FuncType.Goto:
                            m_ScrWork.Instruction = FuncGoto();
                            break;
                        case ScrWork.FuncType.Let:
                            FuncLet();
                            break;
                        case ScrWork.FuncType.Log:
                            FuncLog();
                            break;
                        case ScrWork.FuncType.Exit:
                            complete = true;
                            break;
                        }
                        break;

                    case ScrWork.LineType.Label:
                        lastinst = m_ScrWork.Instruction;
                        // 何もしない
                        break;

                    case ScrWork.LineType.Empty:
                        // 何もしない
                        break;

                    case ScrWork.LineType.SystemLabel:
                        if (m_ScrWork.Count > 0)
                        {
                            // 終了
                            complete = true;
                        }
                        break;

                    default:
                        Debug.Assert(false);
                        break;
                    }

                    m_ScrWork.Instruction++;
                    m_ScrWork.Count++;
                }
            }
            finally
            {
                if (isbreak)
                    m_OnInstruction(lastinst + 1, "-------------------- break");
                else if (complete)
                    m_OnInstruction(lastinst + 2, "-------------------- complete");
                else
                    m_OnInstruction(lastinst + 1, "-------------------- exeption");
            }
        }

        /// <summary>
        /// ●ピクセル色を待つ
        /// </summary>
        private void FuncWaitPixel()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length < 3)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            Match m = m_ScrWork.RegColor.Match(data.Args[2]);
            if (!m.Success)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            Point pt = new Point(Eval(data.Args[0]), Eval(data.Args[1]));
            Color color = Color.FromArgb(unchecked((int)0xFF000000) | Convert.ToInt32(m.Groups[2].Value, 16));
            bool not = (m.Groups[1].Value == "!");
            int timeup = (data.Args.Length >= 4)? Environment.TickCount + Eval(data.Args[3]) : int.MaxValue;

            while (!IsBreak())
            {
                Thread.Sleep(10);

                Color sam;
                if (IsPosColor(pt, color, not, out sam))
                    break;

                if (Environment.TickCount >= timeup)
                {
                    string msg = (data.Args.Length >= 5) ? data.Args[4] : "WaitPixelタイムアウト";
                    msg += string.Format(" ({0:X2}{1:X2}{2:X2})", sam.R, sam.G, sam.B);
                    throw new PlayerException(msg);
                }
            }
        }

        /// <summary>
        /// ●ピクセル色条件付きジャンプ
        /// </summary>
        private int FuncIfGoto(int current)
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 4)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            Match m = m_ScrWork.RegColor.Match(data.Args[2]);
            if (!m.Success)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            Point pt = new Point(Eval(data.Args[0]), Eval(data.Args[1]));
            Color color = Color.FromArgb(unchecked((int)0xFF000000) | Convert.ToInt32(m.Groups[2].Value, 16));
            bool not = (m.Groups[1].Value == "!");

            Color sam;
            if (IsPosColor(pt, color, not, out sam))
            {
                if (m_ScrWork.LabelPoints.ContainsKey(data.Args[3]))
                    return m_ScrWork.LabelPoints[data.Args[3]];

                throw new PlayerException("ラベルが不正です：" + data.Text);
            }

            return current;
        }

        /// <summary>
        /// ●クリック
        /// </summary>
        private void FuncClick()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 2)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            DoClick(new Point(Eval(data.Args[0]), Eval(data.Args[1])));
        }

        /// <summary>
        /// ●マウスカーソル移動
        /// </summary>
        private void FuncMove()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 2)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            DoMove(new Point(Eval(data.Args[0]), Eval(data.Args[1])));
        }

        /// <summary>
        /// ●待つ
        /// </summary>
        private void FuncWait()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 1)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            int timeup = Environment.TickCount + Eval(data.Args[0]);

            while(!IsBreak())
            {
                Thread.Sleep(10);
                if (Environment.TickCount >= timeup)
                    break;
            }
        }

        /// <summary>
        /// ●無条件ジャンプ
        /// </summary>
        private int FuncGoto()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 1)
                throw new PlayerException("引数に誤りがあります：" + data.Text);

            if (m_ScrWork.LabelPoints.ContainsKey(data.Args[0]))
                return m_ScrWork.LabelPoints[data.Args[0]];

            throw new PlayerException("ラベルが不正です：" + data.Text);
        }

        /// <summary>
        /// ●代入
        /// </summary>
        private void FuncLet()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            if (data.Args.Length != 2)
                throw new PlayerException("引数に誤りがあります。");

            m_ScrWork.Vars[data.Args[0]] = Eval(data.Args[1]);
        }

        /// <summary>
        /// ●ログ出し
        /// </summary>
        private void FuncLog()
        {
            ScrWork.LineData data = m_ScrWork.LineDatas[m_ScrWork.Instruction];

            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < data.Args.Length; i++)
            {
                int value = Eval(data.Args[i]);
                sb.Append(string.Format("Log Arg{0} = {1}\r\n", i, value));
            }

            m_OnInstruction(-1, sb.ToString());
        }

        /// <summary>
        /// ●式を評価する
        /// </summary>
        private int Eval(string exp)
        {
            NextToken(exp);
            return Expre();
        }

        /// <summary>
        /// ●式を評価する
        /// </summary>
        private int Expre()
        {
            int val = Term();
            for(;;)
            {
                switch (m_ScrWork.Token)
                {
                case ScrWork.TokenType.Add:
                    NextToken();
                    val += Term();
                    break;
                case ScrWork.TokenType.Sub:
                    NextToken();
                    val -= Term();
                    break;
                default:
                    return val;
                }
            }
        }

        /// <summary>
        /// ●項を評価する
        /// </summary>
        private int Term()
        {
            int val = Factor();
            for(;;)
            {
                switch (m_ScrWork.Token)
                {
                case ScrWork.TokenType.Mul:
                    NextToken();
                    val *= Factor();
                    break;
                case ScrWork.TokenType.Div:
                    NextToken();
                    val /= Factor();
                    break;
                case ScrWork.TokenType.Mod:
                    NextToken();
                    val %= Factor();
                    break;
                default:
                    return val;
                }
            }
        }

        /// <summary>
        /// ●因子を評価する
        /// </summary>
        private int Factor()
        {
            int val;

            switch (m_ScrWork.Token)
            {
            case ScrWork.TokenType.Num:
            case ScrWork.TokenType.Var:
            case ScrWork.TokenType.Sys:
                val = m_ScrWork.Value;
                NextToken();
                return val;

            case ScrWork.TokenType.Lpr:
                NextToken();

                val = Expre();
                if (m_ScrWork.Token != ScrWork.TokenType.Rpr)
                    throw new PlayerException("式に誤りがあります：" + m_ScrWork.LineDatas[m_ScrWork.Instruction].Text);

                NextToken();
                return val;

            default:
                throw new PlayerException("式に誤りがあります：" + m_ScrWork.LineDatas[m_ScrWork.Instruction].Text);
            }
        }

        /// <summary>
        /// ●次のトークンを取得する
        /// </summary>
        private void NextToken(string exp = null)
        {
            m_ScrWork.Tokenizer = (exp != null)? 
                m_ScrWork.RegToken.Match(exp) : m_ScrWork.Tokenizer.NextMatch();

            while (m_ScrWork.Tokenizer.Success)
            {
                if (m_ScrWork.Tokenizer.Groups[6].Success)
                {
                    m_ScrWork.Token = ScrWork.TokenType.End;
                    return;
                }

                string match = m_ScrWork.Tokenizer.Value;
                m_ScrWork.Value = 0;

                switch (match[0])
                {
                case'@':
                    if (!m_ScrWork.Vars.ContainsKey(match.Substring(1)))
                        throw new PlayerException("存在しない変数です：" + match);

                    m_ScrWork.Token = ScrWork.TokenType.Var;
                    m_ScrWork.Value = m_ScrWork.Vars[match.Substring(1)];
                    return;

                case '$':
                    switch (match)
                    {
                    case "$PlayListNum":
                        m_ScrWork.Value = m_PlayList.Count;
                        break;
                    case "$CurOrder":
                        m_ScrWork.Value = m_CurOrder;
                        break;
                    default:
                        throw new PlayerException("存在しない変数です：" + match);
                    }
                    m_ScrWork.Token = ScrWork.TokenType.Sys;
                    return;

                case '+': m_ScrWork.Token = ScrWork.TokenType.Add; return;
                case '-': m_ScrWork.Token = ScrWork.TokenType.Sub; return;
                case '*': m_ScrWork.Token = ScrWork.TokenType.Mul; return;
                case '/': m_ScrWork.Token = ScrWork.TokenType.Div; return;
                case '%': m_ScrWork.Token = ScrWork.TokenType.Mod; return;
                case '(': m_ScrWork.Token = ScrWork.TokenType.Lpr; return;
                case ')': m_ScrWork.Token = ScrWork.TokenType.Rpr; return;

                default:
                    if (m_ScrWork.Tokenizer.Groups[1].Success)
                    {
                        m_ScrWork.Token = ScrWork.TokenType.Num;
                        m_ScrWork.Value = int.Parse(match);
                        return;
                    }
                    if (m_ScrWork.Tokenizer.Groups[4].Success)
                    {
                        // 空白
                        m_ScrWork.Tokenizer = m_ScrWork.Tokenizer.NextMatch();
                        continue;
                    }
                    Debug.Assert(false);
                    break;
                }
                break;
            }

            throw new PlayerException("式に誤りがあります：" + m_ScrWork.LineDatas[m_ScrWork.Instruction].Text);
        }

        /// <summary>
        /// ●スクリプト中断チェック
        /// </summary>
        private bool IsBreak()
        {
            if (m_ScrWork.Section != ScrWork.SystemLabel.Stop && m_Break)
                return true;

            if (m_ScrWork.Section == ScrWork.SystemLabel.WaitEnd && m_Change != ChangeType.None)
                return true;

            return false;
        }

        /// <summary>
        /// ●AmazonMusicを探す
        /// </summary>
        private bool FindAmazonMusic()
        {
            m_AmazonMusic = IntPtr.Zero;
            m_OpeWnd = IntPtr.Zero;

            // Amazon Music メインウィンドウを探す
            m_AmazonMusic = WinApi.FindWindowEx(IntPtr.Zero, IntPtr.Zero, "Amazon Music", null);

            if (m_AmazonMusic == IntPtr.Zero)
                return false;

            // Amazon Music 操作ウィンドウを探す
            m_OpeWnd = WinApi.FindWindowEx(m_AmazonMusic, IntPtr.Zero, "CefBrowserWindow", null);
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            m_OpeWnd = WinApi.FindWindowEx(m_OpeWnd, IntPtr.Zero, "Chrome_WidgetWin_0", null);
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            m_OpeWnd = WinApi.FindWindowEx(m_OpeWnd, IntPtr.Zero, null, "Chrome Legacy Window");
            if (m_OpeWnd == IntPtr.Zero)
                return false;

            return true;
        }

        /// <summary>
        /// ●状態を設定する
        /// </summary>
        private void SetState(ProcState state)
        {
            State = state;
            m_OnStateChange(state);
        }

        /// <summary>
        /// ●指定した位置が指定した色かどうかを調べる
        /// </summary>
        private bool IsPosColor(Point point, Color color, bool not, out Color sam)
        {
            Point pt = ToScreen(point);

            // リモートデスクトップで接続後にハンドル無効例外が出るためリトライする
            try
            {
                for (int i = 0; ; i++)
                {
                    using (Bitmap bmp = new Bitmap(1, 1))
                    {
                        try
                        {
                            using (Graphics g = Graphics.FromImage(bmp))
                            {
                                //g.CopyFromScreen(pt, Point.Empty, bmp.Size); // リモートデスクトップで例外が出るのはコレのせいだった

                                IntPtr hDstDC = IntPtr.Zero;
                                IntPtr hSrcDC = IntPtr.Zero;
                                try
                                {
                                    hDstDC = g.GetHdc();
                                    hSrcDC = WinApi.GetDC(IntPtr.Zero); //WinApi.GetWindowDC(GetOpeWnd());

                                    WinApi.BitBlt(hDstDC, 0, 0, bmp.Width, bmp.Height, hSrcDC, pt.X, pt.Y, WinApi.SRCCOPY);

                                    g.ReleaseHdc(hDstDC);
                                    hDstDC = IntPtr.Zero;

                                    WinApi.ReleaseDC(IntPtr.Zero, hSrcDC);
                                    hSrcDC = IntPtr.Zero;
                                }
                                catch
                                {
                                    if (hDstDC != IntPtr.Zero)
                                    {
                                        g.ReleaseHdc(hDstDC);
                                        hDstDC = IntPtr.Zero;
                                    }
                                    if (hSrcDC != IntPtr.Zero)
                                    {
                                        WinApi.ReleaseDC(IntPtr.Zero, hSrcDC);
                                        hSrcDC = IntPtr.Zero;
                                    }
                                    throw;
                                }

                                sam = bmp.GetPixel(0, 0);
                                bool match = (sam.R == color.R && sam.G == color.G && sam.B == color.B);
                                return (not) ? !match : match;
                            }
                        }
                        catch(Exception ex)
                        {
                            if (i >= 5)
                                throw ex;

                            Thread.Sleep(1);
                        }
                    }
                }
            }
            finally
            {
                GC.Collect();
            }
        }

        /// <summary>
        /// ●クリックする
        /// </summary>
        private void DoClick(Point point)
        {
            if (CursorMove)
                Cursor.Position = ToScreen(point);

            UIntPtr lParam = (UIntPtr)(point.Y << 16 | point.X);
            IntPtr hWnd = GetOpeWnd();
            WinApi.SendMessage(hWnd, WinApi.WM_LBUTTONDOWN, (UIntPtr)WinApi.MK_LBUTTON, lParam);
            Thread.Sleep(100);
            WinApi.SendMessage(hWnd, WinApi.WM_LBUTTONUP, (UIntPtr)0x00000000, lParam);
            Thread.Sleep(100);
        }

        /// <summary>
        /// ●マウス移動
        /// </summary>
        private void DoMove(Point point)
        {
            Cursor.Position = ToScreen(point);
            Thread.Sleep(100);
        }

        /// <summary>
        /// ●次のプレイリストオーダーを設定する
        /// </summary>
        private void SetupNextOrder(ChangeType type)
        {
            switch(type)
            {
            case ChangeType.None:
                m_CurOrder = FindNextOrder();
                break;

            case ChangeType.PrevOrder:    // 現在のジャンルの前のプレイリスト
                m_CurOrder = FindPrevOrder();
                break;
            case ChangeType.NextOrder:    // 現在のジャンルの次のプレイリスト
                m_CurOrder = FindNextOrder();
                break;
            case ChangeType.PrevGenre:    // 前のジャンルの最初のプレイリスト
                m_CurOrder = FindPrevGenre();
                break;
            case ChangeType.NextGenre:    // 次のジャンルの最初のプレイリスト
                m_CurOrder = FindNextGenre();
                break;
            }

            Configuration config = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
            config.AppSettings.Settings["LastOrder"].Value = m_CurOrder.ToString();
            config.AppSettings.Settings["LastGenre"].Value = m_CurGenre.ToString();
            config.Save();
        }

        /// <summary>
        /// ●現在のジャンルの前のプレイリストを取得する
        /// </summary>
        private int FindPrevOrder()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 前のプレイリスト
            List<int> list = new List<int>();
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Genre == m_CurGenre)
                {
                    if (pl.Order != m_CurOrder)
                        list.Add(pl.Order);
                }
            }
            if (list.Count == 0)
                return m_CurOrder;

            if (list.Count == 1)
                return list[0];

            list.Sort();
            list.Reverse();
            foreach (int order in list)
            {
                if (order < m_CurOrder)
                    return order;
            }

            return list[0];
        }

        /// <summary>
        /// ●現在のジャンルの次のプレイリストを取得する
        /// </summary>
        private int FindNextOrder()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のプレイリスト
            List<int> list = new List<int>();
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Genre == m_CurGenre)
                {
                    if (pl.Order != m_CurOrder)
                        list.Add(pl.Order);
                }
            }
            if (list.Count == 0)
                return m_CurOrder; // 無い場合は何もしない

            if (list.Count == 1)
                return list[0];

            list.Sort();
            foreach(int order in list)
            {
                if (order > m_CurOrder)
                    return order;
            }

            return list[0];
        }

        /// <summary>
        /// ●前のジャンルの最初のプレイリストを取得する
        /// </summary>
        private int FindPrevGenre()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のジャンル
            int next = int.MinValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 現在のジャンルより小さい最も大きなジャンル
                if (pl.Genre < m_CurGenre && pl.Genre > next)
                    next = pl.Genre;
            }
            if (next == int.MinValue)
            {
                // 無い場合最も大きなジャンル
                foreach (PlayList pl in m_PlayList)
                {
                    if (pl.Genre > next)
                        next = pl.Genre;
                }
            }
            if (next == int.MinValue)
                return m_CurOrder; // 無い場合は何もしない

            m_CurGenre = next;

            next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 最も小さなオーダー
                if (pl.Genre == m_CurGenre && pl.Order < next)
                    next = pl.Order;
            }

            return next;
        }

        /// <summary>
        /// ●次のジャンルの最初のプレイリストを取得する
        /// </summary>
        private int FindNextGenre()
        {
            // プレイリストが存在しているかを確認する
            if (FixCurrent())
                return m_CurOrder;

            // 次のジャンル
            int next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 現在のジャンルより大きい最も小さなジャンル
                if (pl.Genre > m_CurGenre && pl.Genre < next)
                    next = pl.Genre;
            }
            if (next == int.MaxValue)
            {
                // 無い場合最も小さなジャンル
                foreach (PlayList pl in m_PlayList)
                {
                    if (pl.Genre < next)
                        next = pl.Genre;
                }
            }
            if (next == int.MaxValue)
                return m_CurOrder; // 無い場合は何もしない

            m_CurGenre = next;

            next = int.MaxValue;
            foreach (PlayList pl in m_PlayList)
            {
                // 最も小さなオーダー
                if (pl.Genre == m_CurGenre && pl.Order < next)
                    next = pl.Order;
            }

            return next;
        }

        /// <summary>
        /// ●存在しているプレイリストへ変更する
        /// </summary>
        private bool FixCurrent()
        {
            // プレイリストが存在しているかを確認する
            foreach (PlayList pl in m_PlayList)
            {
                if (pl.Order == m_CurOrder && pl.Genre == m_CurGenre)
                    return false;
            }

            // 存在していない→最初のプレイリストに決定
            m_CurOrder = m_PlayList[0].Order;
            m_CurGenre = m_PlayList[0].Genre;

            return true;
        }

        /// <summary>
        /// ●操作ウィンドウハンドルを取得する
        /// </summary>
        private IntPtr GetOpeWnd()
        {
            try
            {
                // ハンドルの有効性チェック
                if (WinApi.IsWindow(m_OpeWnd))
                {
                    // ウィンドウを最大化する
                    if (!WinApi.IsZoomed(m_AmazonMusic))
                        WinApi.ShowWindow(m_AmazonMusic, WinApi.SW_MAXIMIZE);

                    return m_OpeWnd;
                }
            }
            catch { }

            // 取得し直す
            if (!FindAmazonMusic())
                throw new PlayerException("Amazon Music が見つかりません。");

            // ウィンドウを最大化する
            if (!WinApi.IsZoomed(m_AmazonMusic))
                WinApi.ShowWindow(m_AmazonMusic, WinApi.SW_MAXIMIZE);

            return m_OpeWnd;
        }
    }

    class PlayerException : Exception
    {
        public PlayerException(string msg):base(msg) { }
    }
}
